package main

import (
	"backend/conf"
	"backend/logging"
	"backend/middleware"
	"backend/service"
	"backend/session"
	"database/sql"
	"flag"
	_ "github.com/gin-contrib/cors"
	"github.com/gin-gonic/gin"
	srv "github.com/kardianos/service"
	_ "github.com/mattn/go-sqlite3"
	"log"
	"os"
	"path/filepath"
)
var logger srv.Logger

type Program struct {}

func (p *Program) run() {
	exe, err := os.Executable()
	if err != nil {
		log.Fatal(err)
	}

	cwd := filepath.Dir(exe)

	config := conf.Init(filepath.Join(cwd, "backend.toml"))

	if err := logging.Init(config.GetLogger(cwd)); err != nil {
		log.Panicf("can't init logger: %s\n", err)
	}
	defer logging.Release()

	db, err := sql.Open("sqlite3", filepath.Join(cwd, "backend.db"))
	if err != nil {
		panic(err)
	}

	gin.SetMode(gin.ReleaseMode)
	r := gin.Default()

	//r.Use(cors.Default())

	mgr := session.NewSessionManager(config.SessionKey, config.SessionVal, config.SessionTTL)
	auth := middleware.NewAuth(mgr)

	// login
	u := service.NewUserManage(db, mgr)
	r.POST("/api/login", u.Serve)
	r.PUT("/api/login", auth.Serve, u.Serve)
	r.DELETE("/api/login", auth.Serve, u.Serve)

	// db
	d := service.NewDBManage()
	r.GET("/api/db", auth.Serve, d.Serve)
	r.PUT("/api/db", auth.Serve, d.Serve)

	// server
	s := service.NewServer(d)
	r.GET("/api/server", auth.Serve, s.Retrieve)
	r.POST("/api/server", auth.Serve, s.Serve)
	r.PUT("/api/server", auth.Serve, s.Update)

	r.Static("/static", filepath.Join(cwd, config.Web, "static"))
	r.StaticFile("/", filepath.Join(cwd, config.Web, "index.html"))

	r.Run(config.Address)
}

func (p *Program) Start(s srv.Service) error {
	logger.Info("srv.Start")
	go p.run()
	return nil
}

func (p *Program) Run(s srv.Service) {
	logger.Info("srv.Run")
}

func (p *Program) Stop(s srv.Service) error {
	logger.Info("srv.Stop")
	return nil
}

func main() {
	srvFlag := flag.String("run", "", "control the service")
	flag.Parse()

	srvConf := &srv.Config {
		Name: "httpservermgr",
		DisplayName: "httpserver management",
		Description: "manipulate tophttpserver",
	}

	p := &Program{}
	s, err := srv.New(p, srvConf)
	if err != nil {
		log.Fatal(err)
	}
	logger, err = s.Logger(nil)
	if err != nil {
		log.Fatal(err)
	}

	if len(*srvFlag) != 0 {
		err := srv.Control(s, *srvFlag)
		if err != nil {
			log.Fatalf("error: %v\nvalid actions: %q\n", err, srv.ControlAction)
		}
		return
	}

	err = s.Run()
	if err != nil {
		logger.Error(err)
	}
}
